pub enum LegendScaleType {
    Linear,
    LinearNoOutliers,
    Arithmetic,
    Geometric,
    OverGeometric,
    Harmonic,
    Quantiles,
    StandardisedDescretisation,
    MethodOfBertin,
    NaturalBreaks,
}
